# -*- coding: utf-8 -*-

"""Top-level package for pyspins."""
import pkgutil

__name__ = 'itemset_mining'
__version__ = pkgutil.get_data(__name__, 'VERSION').decode()
__author__ = 'Zax Rosenberg'
__email__ = 'zaxr@protonmail.com'
__copyright__ = '2020'
